import socket
from logging import basicConfig, DEBUG, info, debug, warning, error
from time import sleep

from packaging import version
from requests import get

from dcspy.dcsbios import ProtocolParser
from dcspy.logitech import G13

__version__ = '0.9.0'
basicConfig(format='%(asctime)s | %(levelname)-7s | %(message)s / %(filename)s:%(lineno)d', level=DEBUG)


def attempt_connect(sock: socket.socket) -> None:
    """
    Attempt to connect to localhost.

    :param sock: socket
    """
    connected = False
    info('Waiting for DCS connection...')
    while not connected:
        try:
            sock.connect(('127.0.0.1', 7778))
            info('Connected')
            connected = True
        except socket.error:
            sleep(2)


def check_current_version() -> None:
    """Check if version is current."""
    try:
        url = 'https://api.github.com/repos/emcek/dcspy/releases'
        response = get(url)
        if response.status_code == 200:
            json_response = response.json()
            online_version = json_response['tag_name']
            if version.parse(online_version) > version.parse(__version__):
                info(f'There is new version of dcspy: {online_version}')
            elif version.parse(online_version) == version.parse(__version__):
                info('This is up-to-date version')
            else:
                debug(f'Something goes wrong: local version: {__version__} a online_version: {online_version}')
        else:
            warning(f'Unable to check version online. Try again later. Status={response.status_code}')
    except Exception as e:
        warning(f'Unable to check version online: {e}')


def run() -> None:
    """Main of running function."""
    info(f'dcspy {__version__} https://github.com/emcek/dcspy')
    check_current_version()
    while True:
        parser = ProtocolParser()
        g13 = G13(parser)
        g13.info_display(('G13 initialised OK', 'Waiting for DCS', '', f'dcspy: {__version__}'))

        sock = socket.socket()
        sock.settimeout(None)

        attempt_connect(sock)
        while True:
            try:
                c = sock.recv(1)
                parser.process_byte(c)
                if g13.shouldActivateNewAC:
                    g13.activate_new_ac()
                g13.button_handle(sock)
            except socket.error as exp:
                debug(f'Main loop socket error: {exp}')
                sleep(2)
            except KeyboardInterrupt:
                info('Exit due to Ctrl-C')
                exit(0)
            except Exception as exp:
                error(f'Unexpected error: resetting... {exp.__class__.__name__}', exc_info=True)
                sleep(2)
                break
        del sock
        del g13
        del parser


if __name__ == '__main__':
    run()



from ctypes import CDLL, c_bool, c_wchar_p, c_int, c_ubyte
from itertools import chain

from PIL import Image

# LCD types
TYPE_MONO = 1
TYPE_COLOR = 2

# LCD Monochrome buttons
MONO_BUTTON_0 = 0x00000001
MONO_BUTTON_1 = 0x00000002
MONO_BUTTON_2 = 0x00000004
MONO_BUTTON_3 = 0x00000008

# LCD Color buttons
COLOR_BUTTON_LEFT = 0x00000100
COLOR_BUTTON_RIGHT = 0x00000200
COLOR_BUTTON_OK = 0x00000400
COLOR_BUTTON_CANCEL = 0x00000800
COLOR_BUTTON_UP = 0x00001000
COLOR_BUTTON_DOWN = 0x00002000
COLOR_BUTTON_MENU = 0x00004000

# LCD Monochrome size
MONO_WIDTH = 160
MONO_HEIGHT = 43

# LCD Color size
COLOR_WIDTH = 320
COLOR_HEIGHT = 240

LogiLcdInit = None
LogiLcdIsConnected = None
LogiLcdIsButtonPressed = None
LogiLcdUpdate = None
LogiLcdShutdown = None
LogiLcdMonoSetBackground = None
LogiLcdMonoSetText = None
LogiLcdColorSetBackground = None
LogiLcdColorSetTitle = None
LogiLcdColorSetText = None


def init_dll(dll_path: str) -> None:
    """
    Initialization od dynamic linking library.

    :param dll_path:
    """
    global LogiLcdInit, LogiLcdIsConnected, LogiLcdIsButtonPressed, LogiLcdUpdate, LogiLcdShutdown, LogiLcdMonoSetBackground
    global LogiLcdMonoSetText, LogiLcdColorSetBackground, LogiLcdColorSetTitle, LogiLcdColorSetText

    _dll = CDLL(dll_path)

    # https://docs.python.org/2/library/ctypes.html#fundamental-data-types
    # Generic Functions
    LogiLcdInit = _dll['LogiLcdInit']
    LogiLcdInit.restype = c_bool
    LogiLcdInit.argtypes = (c_wchar_p, c_int)

    LogiLcdIsConnected = _dll['LogiLcdIsConnected']
    LogiLcdIsConnected.restype = c_bool
    LogiLcdIsConnected.argtypes = [c_int]

    LogiLcdIsButtonPressed = _dll['LogiLcdIsButtonPressed']
    LogiLcdIsButtonPressed.restype = c_bool
    LogiLcdIsButtonPressed.argtypes = [c_int]

    LogiLcdUpdate = _dll['LogiLcdUpdate']
    LogiLcdUpdate.restype = None
    # LogiLcdUpdate.argtypes = [None]

    LogiLcdShutdown = _dll['LogiLcdShutdown']
    LogiLcdShutdown.restype = None
    # LogiLcdShutdown.argtypes = [None]

    # Monochrome Lcd Functions
    LogiLcdMonoSetBackground = _dll['LogiLcdMonoSetBackground']
    LogiLcdMonoSetBackground.restype = c_bool
    LogiLcdMonoSetBackground.argtypes = [c_ubyte * 6880]

    LogiLcdMonoSetText = _dll['LogiLcdMonoSetText']
    LogiLcdMonoSetText.restype = c_bool
    LogiLcdMonoSetText.argtypes = (c_int, c_wchar_p)

    # Color LCD Functions
    LogiLcdColorSetBackground = _dll['LogiLcdColorSetBackground']
    LogiLcdColorSetBackground.restype = c_bool
    LogiLcdColorSetBackground.argtypes = [c_ubyte * 307200]

    LogiLcdColorSetTitle = _dll['LogiLcdColorSetTitle']
    LogiLcdColorSetTitle.restype = c_bool
    LogiLcdColorSetTitle.argtypes = (c_wchar_p, c_int, c_int, c_int)

    LogiLcdColorSetText = _dll['LogiLcdColorSetText']
    LogiLcdColorSetText.restype = c_bool
    LogiLcdColorSetText.argtypes = (c_int, c_wchar_p, c_int, c_int, c_int)


def color_gb_picture(im: Image) -> None:
    """
    Set color background picture.

    :param im:
    """
    global LogiLcdColorSetBackground

    LogiLcdColorSetBackground((c_ubyte * 307200)(*list(chain(*list(im.getdata())))))



from ctypes import c_ubyte, sizeof, c_void_p
from importlib import import_module
from logging import basicConfig, DEBUG, info, debug, warning
from math import log2
from platform import architecture
from socket import socket
from sys import maxsize

from PIL import Image, ImageFont, ImageDraw

from dcspy.sdk import lcd_sdk
from dcspy.dcsbios import StringBuffer, ProtocolParser

basicConfig(format='%(asctime)s | %(levelname)-7s | %(message)s / %(filename)s:%(lineno)d', level=DEBUG)


class G13:
    def __init__(self, parser_hook: ProtocolParser) -> None:
        """
        Basic constructor.

        :param parser_hook:
        """
        self.bufferAC = StringBuffer(parser_hook, 0x0000, 16, lambda val: self.set_ac(value=val))
        self.parser = parser_hook
        self.currentAC = ''
        self.currentACHook = None
        self.shouldActivateNewAC = False

        self.isAlreadyPressed = False

        # display parameters
        self.width = lcd_sdk.MONO_WIDTH
        self.height = lcd_sdk.MONO_HEIGHT

        # GLCD Init
        arch = 'x64' if all([architecture()[0] == '64bit', maxsize > 2 ** 32, sizeof(c_void_p) > 4]) else 'x86'
        dll = f"C:\\Program Files\\Logitech Gaming Software\\LCDSDK_8.57.148\\Lib\\GameEnginesWrapper\\{arch}\\LogitechLcdEnginesWrapper.dll"
        lcd_sdk.init_dll(dll)
        lcd_sdk.LogiLcdInit('DCS World', lcd_sdk.TYPE_MONO)

        self.img = Image.new('1', (self.width, self.height), 0)
        self.draw = ImageDraw.Draw(self.img)
        self.font1 = ImageFont.truetype('consola.ttf', 11)
        self.font2 = ImageFont.truetype('consola.ttf', 16)

    def set_ac(self, value: str) -> None:
        """
        Set aircraft.

        :param value:
        """
        if not value == self.currentAC:
            self.currentAC = value
            if value in ('FA-18C_hornet', 'Ka-50', 'F-16C_50'):
                info(f'Detected AC: {value}')
                self.info_display()
                self.shouldActivateNewAC = True
            else:
                # FIXME a może tylko tyo zostawić, żeby po prostu zaczynał aktywaować nowy moduł, a weryfikację zostawić w metodzie poniżej?
                warning(f'Unknown AC data: {value}')
                self.info_display(('Unknown AC data:', self.currentAC))

    def activate_new_ac(self) -> None:
        """Actiate new aircraft."""
        self.shouldActivateNewAC = False
        plane_name = self.currentAC.replace('-', '').replace('_', '')
        plane_class = getattr(import_module('dcspy.aircrafts'), plane_name)
        debug(f'Dynamic load of: {plane_name} as {self.currentAC}')
        self.currentACHook = plane_class(self)

    def info_display(self, message=('', '')) -> None:
        """
        Send message to display.

        :param message:
        """
        # clear bitmap
        self.draw.rectangle((0, 0, self.width, self.height), 0, 0)
        # self.ClearDisplay()

        if message == ('', ''):
            self.draw.text((0, 0), self.currentAC, 1, self.font1)
        else:
            y = 0
            # self.draw.text((0,0), message, 1, self.font1)
            for line in message:
                self.draw.text((0, y), line, 1, self.font1)
                y = y + 10

        self.update_display(self.img)

    def update_display(self, img: Image) -> None:
        """
        Update display.

        :param img:
        """
        pixels = list(img.getdata())
        for i, _ in enumerate(pixels):
            pixels[i] *= 128

        # put bitmap array into display
        if lcd_sdk.LogiLcdIsConnected(lcd_sdk.TYPE_MONO):
            lcd_sdk.LogiLcdMonoSetBackground((c_ubyte * (self.width * self.height))(*pixels))
            lcd_sdk.LogiLcdUpdate()
        else:
            warning('LCD is not connected')

    def clear_display(self, true_clear=False) -> None:
        """
        Clear display.

        :param true_clear:
        """
        lcd_sdk.LogiLcdMonoSetBackground((c_ubyte * (self.width * self.height))(*[0] * (self.width * self.height)))
        if true_clear:
            for i in range(4):
                lcd_sdk.LogiLcdMonoSetText(i, '')
        lcd_sdk.LogiLcdUpdate()

    def check_buttons(self) -> int:
        """
        Check button state.

        :return:
        """
        for btn in (lcd_sdk.MONO_BUTTON_0, lcd_sdk.MONO_BUTTON_1, lcd_sdk.MONO_BUTTON_2, lcd_sdk.MONO_BUTTON_3):
            if lcd_sdk.LogiLcdIsButtonPressed(btn):
                if not self.isAlreadyPressed:
                    self.isAlreadyPressed = True
                    return int(log2(btn)) + 1
                return 0
        self.isAlreadyPressed = False
        return 0

    def button_handle(self, s: socket) -> None:
        """
        Button handler.

        :param s:
        """
        button = self.check_buttons()
        if button:
            s.send(bytes(self.currentACHook.button_handle_specific_ac(button), 'utf-8'))



from logging import basicConfig, DEBUG, debug

from dcspy.dcsbios import StringBuffer

basicConfig(format='%(asctime)s | %(levelname)-7s | %(message)s / %(filename)s:%(lineno)d', level=DEBUG)


class AircraftHandler:
    def __init__(self, display_handler) -> None:
        """
        Basic constructor.

        :param display_handler:
        :type display_handler: G13Handler
        """
        self.g13 = display_handler

    def button_handle_specific_ac(self, button_pressed: int) -> str:
        """
        Button handler for spacific aircraft.

        :param button_pressed:
        :return:
        """
        debug(f'{self.__class__.__name__} Button: {button_pressed}')
        return '\n'

    def update_display(self) -> None:
        """Update display."""
        self.g13.draw.rectangle((0, 0, self.g13.width, self.g13.height), 0, 0)  # clear bitmap

    def set_data(self, selector: str, value: str, update=True) -> None:
        """
        Set new data.

        :param selector:
        :param value:
        :param update:
        """
        setattr(self, selector, value)
        debug(f'value: {value}')
        if update:
            self.update_display()


class FA18Chornet(AircraftHandler):
    def __init__(self, display_handler) -> None:
        """
        Basic constructor.

        :param display_handler:
        :type display_handler: G13Handler
        """
        super().__init__(display_handler)
        self.ScratchpadStr1 = ''
        self.ScratchpadStr2 = ''
        self.ScratchpadNum = ''
        self.OptionDisplay1 = ''
        self.OptionDisplay2 = ''
        self.OptionDisplay3 = ''
        self.OptionDisplay4 = ''
        self.OptionDisplay5 = ''
        self.COMM1 = ''
        self.COMM2 = ''
        self.OptionCueing1 = ''
        self.OptionCueing2 = ''
        self.OptionCueing3 = ''
        self.OptionCueing4 = ''
        self.OptionCueing5 = ''
        self.FuelTotal = ''

        self.bufferScratchpadStr1 = StringBuffer(self.g13.parser, 0x744e, 2, lambda s: self.set_data('ScratchpadStr1', s))
        self.bufferScratchpadStr2 = StringBuffer(self.g13.parser, 0x7450, 2, lambda s: self.set_data('ScratchpadStr2', s))
        self.bufferScratchpadNum = StringBuffer(self.g13.parser, 0x7446, 8, lambda s: self.set_data('ScratchpadNum', s))
        self.bufferOptionDisplay1 = StringBuffer(self.g13.parser, 0x7432, 4, lambda s: self.set_data('OptionDisplay1', s))
        self.bufferOptionDisplay2 = StringBuffer(self.g13.parser, 0x7436, 4, lambda s: self.set_data('OptionDisplay2', s))
        self.bufferOptionDisplay3 = StringBuffer(self.g13.parser, 0x743a, 4, lambda s: self.set_data('OptionDisplay3', s))
        self.bufferOptionDisplay4 = StringBuffer(self.g13.parser, 0x743e, 4, lambda s: self.set_data('OptionDisplay4', s))
        self.bufferOptionDisplay5 = StringBuffer(self.g13.parser, 0x7442, 4, lambda s: self.set_data('OptionDisplay5', s))
        self.bufferCOMM1 = StringBuffer(self.g13.parser, 0x7424, 2, lambda s: self.set_data('COMM1', s))
        self.bufferCOMM2 = StringBuffer(self.g13.parser, 0x7426, 2, lambda s: self.set_data('COMM2', s))
        self.bufferOptionCueing1 = StringBuffer(self.g13.parser, 0x7428, 1, lambda s: self.set_data('OptionCueing1', s))
        self.bufferOptionCueing2 = StringBuffer(self.g13.parser, 0x742a, 1, lambda s: self.set_data('OptionCueing2', s))
        self.bufferOptionCueing3 = StringBuffer(self.g13.parser, 0x742c, 1, lambda s: self.set_data('OptionCueing3', s))
        self.bufferOptionCueing4 = StringBuffer(self.g13.parser, 0x742e, 1, lambda s: self.set_data('OptionCueing4', s))
        self.bufferOptionCueing5 = StringBuffer(self.g13.parser, 0x7430, 1, lambda s: self.set_data('OptionCueing5', s))
        self.bufferFuelTotal = StringBuffer(self.g13.parser, 0x748a, 6, lambda s: self.set_data('FuelTotal', s))

    def update_display(self) -> None:
        """Update display."""
        super().update_display()

        # Scrachpad
        self.g13.draw.text((0, 0),
                           self.ScratchpadStr1 + self.ScratchpadStr2 + self.ScratchpadNum,
                           1, self.g13.font2)
        self.g13.draw.line((0, 20, 115, 20), 1, 1)

        # comm1
        self.g13.draw.rectangle((0, 29, 20, 42), 0, 1)
        self.g13.draw.text((2, 29), self.COMM1, 1, self.g13.font2)

        # comm2
        offset_comm2 = 44
        self.g13.draw.rectangle((139 - offset_comm2, 29, 159 - offset_comm2, 42), 0, 1)
        self.g13.draw.text((140 - offset_comm2, 29), self.COMM2, 1, self.g13.font2)

        # option display 1..5 with cueing
        for i in range(1, 6):
            offset = (i - 1) * 8
            self.g13.draw.text((120, offset),
                               f'{i}{getattr(self, f"OptionCueing{i}")}{getattr(self, f"OptionDisplay{i}")}',
                               1, self.g13.font1)

        # Fuel Totaliser
        self.g13.draw.text((36, 29), self.FuelTotal, 1, self.g13.font2)

        self.g13.update_display(self.g13.img)

    def set_data(self, selector: str, value: str, update=True) -> None:
        """
        Set new data.

        :param selector:
        :param value:
        :param update:
        """
        if selector in ('ScratchpadStr1', 'ScratchpadStr2', 'COMM1', 'COMM2'):
            value = value.replace('`', '1').replace('~', '2')
        super().set_data(selector, value, update)

    def button_handle_specific_ac(self, button_pressed: int) -> str:
        """
        Button handler for spacific aircraft.

        :param button_pressed:
        :return:
        """
        action = {1: 'UFC_COMM1_CHANNEL_SELECT DEC',
                  2: 'UFC_COMM1_CHANNEL_SELECT INC',
                  3: 'UFC_COMM2_CHANNEL_SELECT DEC',
                  4: 'UFC_COMM2_CHANNEL_SELECT INC'}
        debug(f'{self.__class__.__name__} Button: {button_pressed}')
        debug(f'Request: {action[button_pressed]}')
        return f'{action[button_pressed]}\n'


class F16C50(AircraftHandler):
    def __init__(self, display_handler) -> None:
        """
        Basic constructor.

        :param display_handler:
        :type display_handler: G13Handler
        """
        super().__init__(display_handler)
        self.DEDLine1 = ''
        self.DEDLine2 = ''
        self.DEDLine3 = ''
        self.DEDLine4 = ''
        self.DEDLine5 = ''

        self.bufferDEDLine1 = StringBuffer(self.g13.parser, 0x44fc, 50, lambda s: self.set_data('DEDLine1', s))
        self.bufferDEDLine2 = StringBuffer(self.g13.parser, 0x452e, 50, lambda s: self.set_data('DEDLine2', s))
        self.bufferDEDLine3 = StringBuffer(self.g13.parser, 0x4560, 50, lambda s: self.set_data('DEDLine3', s))
        self.bufferDEDLine4 = StringBuffer(self.g13.parser, 0x4592, 50, lambda s: self.set_data('DEDLine4', s))
        self.bufferDEDLine5 = StringBuffer(self.g13.parser, 0x45c4, 50, lambda s: self.set_data('DEDLine5', s))

    def update_display(self) -> None:
        """Update display."""
        super().update_display()

        for i in range(1, 6):
            offset = (i - 1) * 8
            self.g13.draw.text((0, offset), getattr(self, f'DEDLine{i}'), 1, self.g13.font1)
        self.g13.update_display(self.g13.img)


class Ka50(AircraftHandler):
    def __init__(self, display_handler):
        """
        Basic constructor.

        :param display_handler:
        :type display_handler: G13Handler
        """
        super().__init__(display_handler)
        self.l1_apostr1 = ''
        self.l1_apostr2 = ''
        self.l1_point = ''
        self.l1_sign = ''
        self.l1_text = ''
        self.l2_apostr1 = ''
        self.l2_apostr2 = ''
        self.l2_point = ''
        self.l2_sign = ''
        self.l2_text = ''

        self.buffer_l1_apostr1 = StringBuffer(self.g13.parser, 0x1934, 1, lambda s: self.set_data('l1_apostr1', s))
        self.buffer_l1_apostr2 = StringBuffer(self.g13.parser, 0x1936, 1, lambda s: self.set_data('l1_apostr2', s))
        self.buffer_l1_point = StringBuffer(self.g13.parser, 0x1930, 1, lambda s: self.set_data('l1_point', s))
        self.buffer_l1_sign = StringBuffer(self.g13.parser, 0x1920, 1, lambda s: self.set_data('l1_sign', s))
        self.buffer_l1_text = StringBuffer(self.g13.parser, 0x1924, 6, lambda s: self.set_data('l1_text', s))
        self.buffer_l2_apostr1 = StringBuffer(self.g13.parser, 0x1938, 1, lambda s: self.set_data('l2_apostr1', s))
        self.buffer_l2_apostr2 = StringBuffer(self.g13.parser, 0x193a, 1, lambda s: self.set_data('l2_apostr2', s))
        self.buffer_l2_point = StringBuffer(self.g13.parser, 0x1932, 1, lambda s: self.set_data('l2_point', s))
        self.buffer_l2_sign = StringBuffer(self.g13.parser, 0x1922, 1, lambda s: self.set_data('l2_sign', s))
        self.buffer_l2_text = StringBuffer(self.g13.parser, 0x192a, 6, lambda s: self.set_data('l2_text', s))

    def button_handle_specific_ac(self, button_pressed: int) -> str:
        """
        Button handler for spacific aircraft.

        :param button_pressed:
        :return:
        """
        debug(f'{self.__class__.__name__} Button: {button_pressed}')
        return '\n'

    def update_display(self) -> None:
        """Update display."""
        super().update_display()
        text1, text2 = '', ''
        if self.l1_text:
            text1 = f' {self.l1_text[-5:-3]}{self.l1_apostr1}{self.l1_text[-3:-1]}{self.l1_apostr2}{self.l1_text[-1]}'
        if self.l2_text:
            text2 = f'{self.l2_text[-6:-3]}{self.l2_apostr1}{self.l2_text[-3:-1]}{self.l2_apostr2}{self.l2_text[-1]}'
        line1 = f'{self.l1_sign} {text1} {self.l1_point}'
        line2 = f'{self.l2_sign} {text2} {self.l2_point}'
        self.g13.draw.text((0, 0), line1, 1, self.g13.font1)
        self.g13.draw.text((0, 8), line2, 1, self.g13.font1)
        self.g13.update_display(self.g13.img)



from struct import pack
from typing import Callable, Set


class ProtocolParser:
    def __init__(self) -> None:
        """Basic constructor."""
        self.__state = 'WAIT_FOR_SYNC'
        self.__sync_byte_count = 0
        self.__address = 0
        self.__count = 0
        self.__data = 0
        self.write_callbacks: Set[Callable] = set()
        self.frame_sync_callbacks: Set[Callable] = set()

    def process_byte(self, b: bytes) -> None:
        """
        Process byte.

        :param b:
        """
        c = ord(b)
        if self.__state == 'ADDRESS_LOW':
            self.__address = c
            self.__state = 'ADDRESS_HIGH'
        elif self.__state == 'ADDRESS_HIGH':
            self.__address += c * 256
            if self.__address != 0x5555:
                self.__state = 'COUNT_LOW'
            else:
                self.__state = 'WAIT_FOR_SYNC'
        elif self.__state == 'COUNT_LOW':
            self.__count = c
            self.__state = 'COUNT_HIGH'
        elif self.__state == 'COUNT_HIGH':
            self.__count += 256 * c
            self.__state = 'DATA_LOW'
        elif self.__state == 'DATA_LOW':
            self.__data = c
            self.__count -= 1
            self.__state = 'DATA_HIGH'
        elif self.__state == 'DATA_HIGH':
            self.__data += 256 * c
            self.__count -= 1
            for callback in self.write_callbacks:
                callback(self.__address, self.__data)
            self.__address += 2
            if self.__count == 0:
                self.__state = 'ADDRESS_LOW'
            else:
                self.__state = 'DATA_LOW'

        if c == 0x55:
            self.__sync_byte_count += 1
        else:
            self.__sync_byte_count = 0

        if self.__sync_byte_count == 4:
            self.__state = 'ADDRESS_LOW'
            self.__sync_byte_count = 0
            for callback in self.frame_sync_callbacks:
                callback()


class StringBuffer:
    def __init__(self, parser: ProtocolParser, address: int, length: int, callback: Callable) -> None:
        """
        Basic constructor.

        :param parser:
        :param address:
        :param length:
        :param callback:
        """
        self.__address = address
        self.__length = length
        self.__dirty = False
        self.buffer = bytearray(length)
        self.callbacks: Set[Callable] = set()
        if callback:
            self.callbacks.add(callback)
        parser.write_callbacks.add(lambda addr, data: self.on_dcsbios_write(address=addr, data=data))

    def set_char(self, i, c) -> None:
        """
        Set char.

        :param i:
        :param c:
        """
        if self.buffer[i] != c:
            self.buffer[i] = c
            self.__dirty = True

    def on_dcsbios_write(self, address: int, data: int) -> None:
        """
        Callback function.

        :param address:
        :param data:
        """
        if self.__address <= address < self.__address + self.__length:
            data_bytes = pack('<H', data)
            self.set_char(address - self.__address, data_bytes[0])
            if self.__address + self.__length > (address + 1):
                self.set_char(address - self.__address + 1, data_bytes[1])

        if address == 0xfffe and self.__dirty:
            self.__dirty = False
            s = self.buffer.split(b'\x00')[0].decode('latin-1')
            for callback in self.callbacks:
                callback(s)


class IntegerBuffer:
    def __init__(self, parser: ProtocolParser, address: int, mask: int, shift_by: int, callback: Callable) -> None:
        """
        Basic constructor.

        :param parser:
        :param address:
        :param mask:
        :param shift_by:
        :param callback:
        """
        self.__address = address
        self.__mask = mask
        self.__shift_by = shift_by
        self.__value = None
        self.callbacks: Set[Callable] = set()
        if callback:
            self.callbacks.add(callback)
        parser.write_callbacks.add(lambda addr, data: self.on_dcsbios_write(address=addr, data=data))

    def on_dcsbios_write(self, address: int, data: int) -> None:
        """
        Callback function.

        :param address:
        :param data:
        """
        if address == self.__address:
            value = (data & self.__mask) >> self.__shift_by
            if self.__value != value:
                self.__value = value
                for callback in self.callbacks:
                    callback(value)
